"""
Wrapper for image files.

OCR:  https://pypi.org/project/pytesseract/
EXIF: https://blog.matthewgove.com/2022/05/13/how-to-bulk-edit-your-photos-exif-data-with-10-lines-of-python/
Tags: https://exiftool.org/TagNames/EXIF.html
"""
import io
import re
from pathlib import Path
from typing import Optional, Union

import pytesseract
from PIL import Image, TiffImagePlugin
from PIL.ExifTags import TAGS
from rich.pretty import pprint

from clown_sort.config import Config
from clown_sort.filename_extractor import FilenameExtractor
from clown_sort.files.sortable_file import RuleMatch, SortableFile
from clown_sort.util.filesystem_helper import copy_file_creation_time
from clown_sort.util.logging import log
from clown_sort.util.rich_helper import console, error_text, warning_text

THUMBNAIL_DIMENSIONS = (512, 512)
IMAGE_DESCRIPTION = 'ImageDescription'
FILENAME_LENGTH_TO_CONSIDER_SORTED = 80

EXIF_CODES = {
    IMAGE_DESCRIPTION: 270,
}


class ImageFile(SortableFile):
    def copy_file_to_sorted_dir(self, destination_path: Path, match: Optional[re.Match] = None) -> None:
        """
        Copies to a new file and injects the ImageDescription exif tag.
        If :destination_subdir is given new file will be in :destination_subdir off
        of the configured :destination_dir. Returns new file path.
        """
        exif_data = self.raw_exif_dict()
        self._log_copy_file(destination_path, match)

        if self.extracted_text() is not None:
            exif_data.update([(EXIF_CODES[IMAGE_DESCRIPTION], self.extracted_text())])

        if Config.dry_run:
            return

        try:
            self.pillow_image_obj().save(destination_path, exif=exif_data)
            copy_file_creation_time(self.file_path, destination_path)
        except (NotImplementedError, TypeError, ValueError) as e:
            console.print_exception()
            console.print(error_text(f"Failed to save '{self.file_path}'").append("\nEXIF DATA:"))
            pprint(self.printable_exif_dict(), expand_all=True, indent_guides=False)

            if self.extname.lower().startswith('.tif'):
                console.print(error_text("TIFF files are sometimes not supported."))

            raise e

    def new_basename(self) -> str:
        """Return a descriptive string usable in a filename."""
        if self._new_basename is not None:
            return self._new_basename

        if self.extracted_text() is None \
                or len(self.extracted_text()) == 0 \
                or len(self.basename) > FILENAME_LENGTH_TO_CONSIDER_SORTED:
            self._new_basename = self.basename
        else:
            self._filename_extractor = FilenameExtractor(self)
            self._new_basename = self._filename_extractor.filename()

        self._new_basename = self._new_basename.replace('""', '"')
        return self._new_basename

    def thumbnail_bytes(self) -> bytes:
        """Return bytes for a thumbnail."""
        image = self.pillow_image_obj()
        image.thumbnail(THUMBNAIL_DIMENSIONS)
        _thumbnail_bytes = io.BytesIO()
        image.save(_thumbnail_bytes, format="PNG")
        return _thumbnail_bytes.getvalue()

    def extracted_text(self) -> Optional[str]:
        """Use Tesseract to OCR the text in the image, which is returned as a string."""
        if self.text_extraction_attempted:
            return self._extracted_text

        self._extracted_text = ImageFile.ocr_text(self.pillow_image_obj(), str(self.file_path))
        self.text_extraction_attempted = True
        return self._extracted_text

    def exif_dict(self) -> dict:
        """Return a key/value list of exif tags where keys are strings."""
        raw_exif_tags = self.pillow_image_obj().getexif()
        return {TAGS[k]: v for k,v in raw_exif_tags.items()}

    def raw_exif_dict(self) -> Image.Exif:
        """Return a key/value list of exif tags where keys are integers."""
        return self.pillow_image_obj().getexif()

    def printable_exif_dict(self, exif_data: Optional[dict] = None) -> dict:
        """Return a dict of the exif tags with the unprintable values coerced to something printable."""
        pprint(f"called with exif_data: {exif_data}")
        exif_data = exif_data or self.exif_dict()
        new_dict = {}

        for k, v in exif_data.items():
            if isinstance(v, bytes):
                new_dict[k] = f"<{len(v)} raw bytes>"
            elif isinstance(v, TiffImagePlugin.IFDRational):
                new_dict[k] = float(v)
            else:
                new_dict[k] = v

        return new_dict

    def pillow_image_obj(self) -> Image.Image:
        """Return the file as Pillow Image object."""
        return Image.open(self.file_path)

    def can_be_presented_in_popup(self) -> bool:
        return True

    def __repr__(self) -> str:
        return f"ImageFile('{self.file_path}')"

    # def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
    #     super().__rich_console__(console, options)
    #     log.debug(f"RAW EXIF: {self.raw_exif_dict()}")

    @staticmethod
    def ocr_text(image: Image.Image, image_name: str) -> Optional[str]:
        """Use pytesseract to OCR the text in the image and return it as a string."""
        text = None

        try:
            text = pytesseract.image_to_string(image)
        except pytesseract.pytesseract.TesseractError as e:
            console.print_exception()
            console.print(warning_text(f"Tesseract OCR failure '{image_name}'! No OCR text extracted..."))
        except OSError as e:
            if 'truncated' in str(e):
                console.print(warning_text(f"Truncated image file '{image_name}'!"))
            else:
                console.print_exception()
                console.print(f"Error while extracting '{image_name}'!", style='bright_red')
                raise e
        except Exception as e:
            console.print_exception()
            console.print(f"Error while extracting '{image_name}'!", style='bright_red')
            raise e

        return None if text is None else text.strip()



"""
Class to contain a text sorting rule.
"""
import csv
import importlib.resources
import re
from dataclasses import dataclass
from os import environ
from pathlib import Path
from typing import List, Optional, Union

from clown_sort.util.constants import CRYPTO, PACKAGE_NAME
from clown_sort.util.rich_helper import print_error

RULES_CSV_PATHS = 'RULES_CSV_PATHS'
SORTING_RULES_DIR = importlib.resources.files(PACKAGE_NAME).joinpath('sorting_rules')
CRYPTO_RULES_CSV_PATH = Path(str(SORTING_RULES_DIR.joinpath('crypto.csv')))


class SortRuleParseError(RuntimeError):
    pass


@dataclass
class SortRule:
    folder: str
    regex: Union[str, re.Pattern]

    def __post_init__(self):
        try:
            self.regex = re.compile(self.regex, re.IGNORECASE | re.MULTILINE)
        except re.error as e:
            msg = f"{str(e)} while processing '{self.folder}' sort rule '{self.regex}'"
            raise SortRuleParseError(msg) from e

    @classmethod
    def load_rules_csv(cls, file_path: Path) -> List['SortRule']:
        """Turn a CSV of sort rules into a list of SortRule objects."""
        sort_rules = []

        with open(Path(file_path), mode='r') as csvfile:
            for row in csv.DictReader( filter(SortRule.is_valid_row, csvfile), delimiter=','):
                try:
                    sort_rules.append(cls(row['folder'], row['regex']))
                except SortRuleParseError as e:
                    print_error(f"{str(e)} in file '{file_path}'!")
                    raise e

        return sort_rules

    @staticmethod
    def default_rules_csv_paths() -> List[Path]:
        """Check for RULES_CSV_PATHS in env vars otherwise return default crypto rules."""
        return SortRule.sort_rules_csvs(environ.get(RULES_CSV_PATHS, CRYPTO_RULES_CSV_PATH))

    @staticmethod
    def is_valid_row(row: str) -> bool:
        """Returns false for whitespace rows and those that start with '#'."""
        stripped_row = row.strip()
        return len(stripped_row) > 0 and stripped_row[0] != '#'

    @staticmethod
    def sort_rules_csvs(csv_paths: Optional[str]) -> List[Path]:
        """Expects str of paths separated by ':'. Replace the string 'crypto' with the default CSV."""
        if isinstance(csv_paths, Path):
            return [csv_paths]
        elif csv_paths is None:
            return SortRule.default_rules_csv_paths()

        return [
            CRYPTO_RULES_CSV_PATH if csv_path == CRYPTO else Path(csv_path)
            for csv_path in csv_paths.split(':')
        ]

    def __eq__(self, other: 'SortRule') -> bool:
        return self.folder == other.folder and self.regex == other.regex



import pdfplumber

PDF = '/Users/syblius/Pictures/Screenshots/Sorted/Tether/Tether Portfolio Report 2021-03-31 from Ansbacher.pdf'

pdf = pdfplumber.open(PDF)
pdf.pages[5].to_image()

table_settings = {
    "vertical_strategy": "text",
    "horizontal_strategy": "text",
    # "snap_y_tolerance": 5,
    # "intersection_x_tolerance": 15,
}

table = pdf.pages[5].extract_table(table_settings)



"""
Functions and constants having to do with the filesystem.
importlib explanation: https://fossies.org/linux/Python/Lib/importlib/resources.py
"""
import os
import re
import stat
import time
from datetime import datetime
from getpass import getuser
from os import path
from pathlib import Path
from typing import List, Optional, Union

from filedate.Utils import Copy
from filedate import File
from rich.console import Console
from rich.text import Text

from clown_sort.util.constants import MAC_SCREENSHOT_REGEX, SCREENSHOT_REGEX
from clown_sort.util.logging import log
from clown_sort.util.string_helper import spaces_to_underscores

PDF_EXTENSION = '.pdf'
MOVIE_FILE_EXTENSIONS = ['.mov', '.flv', '.avi']
IMAGE_FILE_EXTENSIONS = [f".{ext}" for ext in 'tiff jpg jpeg png heic'.split()]
SORTABLE_FILE_EXTENSIONS = IMAGE_FILE_EXTENSIONS + [PDF_EXTENSION, '.mov']
MAC_SCREENSHOT_TIMESTAMP_FORMAT = '%Y-%m-%d at %I.%M.%S %p'

ANONYMIZED_USERNAME = 'uzor'
CURRENT_USERNAME = getuser()


def files_in_dir(dir: Union[os.PathLike, str], with_extname: Optional[str] = None) -> List[str]:
    """Paths for non-hidden, non-directory files, optionally ending in 'with_extname'."""
    files = [file for file in _non_hidden_files_in_dir(dir) if not path.isdir(file)]

    if with_extname:
        files = [f for f in files if f.endswith(f".{with_extname}")]

    return files


def subdirs_of_dir(dir: Union[os.PathLike, str]) -> List[str]:
    """Find non-hidden subdirs in 'dir'."""
    return [file for file in _non_hidden_files_in_dir(dir) if path.isdir(file)]


def timestamp_for_filename() -> str:
    """Returns a string showing current time in a file name friendly format."""
    return datetime.now().strftime("%Y-%m-%dT%H.%M.%S")


def strip_bad_chars(text: str) -> str:
    """Remove chars that don't work well in filenames."""
    text = ' '.join(text.splitlines()).replace('\\s+', ' ')
    text = re.sub('’', "'", text).replace('|', 'I').replace(',', ',')
    return re.sub('[^-0-9a-zA-Z@.,?_:=#\'\\$" ()]+', '_', text).replace('  ', ' ')


def strip_mac_screenshot(text: str) -> str:
    """Strip default macOS screenshot format from filename."""
    return re.sub(SCREENSHOT_REGEX, '', text).strip()


def is_image(file_path: Union[str, Path]) -> bool:
    return Path(file_path).suffix in IMAGE_FILE_EXTENSIONS


def is_movie(file_path: Union[str, Path]) -> bool:
    return Path(file_path).suffix in MOVIE_FILE_EXTENSIONS


def is_pdf(file_path: Union[str, Path]) -> bool:
    return Path(file_path).suffix == PDF_EXTENSION


def is_sortable(file_path: Union[str, Path]) -> bool:
    """Return True if it's a file clown_sort knows how to process."""
    return Path(file_path).suffix in SORTABLE_FILE_EXTENSIONS


def insert_suffix_before_extension(file_path: Path, suffix: str, separator: str = '__') -> Path:
    """Inserting 'page 1' suffix in 'path/to/file.jpg' -> '/path/to/file__page_1.jpg'."""
    suffix = spaces_to_underscores(strip_bad_chars(suffix))
    file_path_without_extension = file_path.with_suffix('')
    return Path(f"{file_path_without_extension}{separator}{suffix}{file_path.suffix}")


def create_dir_if_it_does_not_exist(dir: Path) -> None:
    """Like it says on the tin."""
    if dir.exists():
        return

    log.warning(f"Need to create '{dir}'")
    dir.mkdir(parents=True, exist_ok=True)


def copy_file_creation_time(source_file: Path, destination_file: Path) -> None:
    """Copy the file creation timestamp from source_file to destination_file."""
    try:
        Copy(str(source_file), str(destination_file)).all()
    except FileNotFoundError:
        msg = Text("WARNING! couldn't copy file creation timestamp because file does not exist for ")
        msg.append(f"'{destination_file}'!")
        Console().print(msg, style='bright_yellow')
        return

    _set_permissions(destination_file)


def set_timestamp_based_on_screenshot_filename(file_path: Path) -> None:
    """Infer a timestamp based on the filename and then change the 'Last Modified' property to match."""
    file_timestamp = extract_timestamp_from_filename(str(file_path))
    print(f"Parsed {file_timestamp} from '{file_path.name}'")
    print("    last modified: %s" % time.ctime(os.path.getmtime(file_path)))
    print("    created: %s" % time.ctime(os.path.getctime(file_path)))
    file_date = File(file_path)
    file_date.set(created = file_timestamp, modified = file_timestamp)
    _set_permissions(file_path)


def extract_timestamp_from_filename(filename: str) -> datetime:
    """Infer a timestamp based on the filename. Assumes there is an iso8601 section in filename."""
    filename = os.path.basename(filename)
    match = MAC_SCREENSHOT_REGEX.match(filename)

    if not match:
        raise ValueError(f"'{filename}' is not a timestamped screenshot file")

    return datetime.strptime(match.group(1), MAC_SCREENSHOT_TIMESTAMP_FORMAT)


def loggable_filename(file_path: Path, config: 'Config') -> str:
    """Return a loggable version of the filename."""
    if config.hide_dirs:
        return Path(file_path).name

    filename = str(file_path)

    if config.anonymize_user_dir:
        username_str = f"{path.sep}{CURRENT_USERNAME}{path.sep}"
        anonuser_str = f"{path.sep}{ANONYMIZED_USERNAME}{path.sep}"
        return filename.replace(username_str, anonuser_str)
    else:
        return filename


def _non_hidden_files_in_dir(dir: Union[os.PathLike, str]) -> List[str]:
    return [path.join(dir, file) for file in os.listdir(dir) if not file.startswith('.')]


def _set_permissions(file_path: Path) -> None:
    """The filedate library has a bad habit of changing all the permissions so we change them back."""
    os.chmod(file_path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)



"""
Open a GUI window to allow manual name / select.
TODO: rename to something more appropriate
"""
import sys
from os import path, remove
from typing import Union

from clown_sort.config import Config
from clown_sort.filename_extractor import FilenameExtractor
from clown_sort.util.logging import log
from clown_sort.util.rich_helper import bullet_text, console, indented_bullet
from clown_sort.util.string_helper import is_empty

RADIO_COLS = 11
SELECT_SIZE = 45
DELETE = 'Delete'
OK = 'Move'
OPEN = 'Preview File'
SKIP = 'Skip'
EXIT = 'Exit'


def process_file_with_popup(image: Union['ImageFile', 'PdfFile']) -> None:
    # Do the import here so as to allow usage without installing PySimpleGUI
    import FreeSimpleGUI as psg
    psg.theme('SystemDefault1')
    suggested_filename = FilenameExtractor(image).filename()
    sort_dirs = [path.basename(dir) for dir in Config.get_sort_dirs()]
    max_dirname_length = max([len(dir) for dir in sort_dirs])
    thumbnail_bytes = image.thumbnail_bytes()

    if thumbnail_bytes is None:
        log.warn(f"Failed to get a thumbnail; skipping...")
        return

    layout = [
        [psg.Column([[psg.Image(data=thumbnail_bytes, key="-IMAGE-")]], justification='center')],
        [psg.HSep()],
        [psg.Text("Choose Filename:")],
        [psg.Input(suggested_filename, size=(len(suggested_filename), 1))],# font=("Courier New", 12))],
        [
            psg.Text(f"Choose Directory:"),
            psg.Combo(sort_dirs, size=(max_dirname_length, SELECT_SIZE)),
            psg.Text(f"(Enter custom text to create new directory. If no directory is chosen file will be copied to '{Config.sorted_screenshots_dir}'.)")
        ],
        [
            psg.Button(OK, bind_return_key=True),
            psg.Button(DELETE),
            psg.Button(OPEN),
            psg.Button(SKIP),
            psg.Button(EXIT)
        ]
    ]

    window = psg.Window(image.basename, layout)

    # Event Loop
    while True:
        event, values = window.Read()

        if event == OPEN:
            image.preview()
            continue

        window.close()

        if event == DELETE:
            log.warning(f"Deleting '{image.file_path}'")
            remove(image.file_path)
            return
        elif event == SKIP or event == psg.WIN_CLOSED:
            return
        elif event == EXIT:
            sys.exit()
        elif event == OK:
            break

    log.debug(f"All values: {values}")
    chosen_filename = values[1]
    new_subdir = values[2]
    destination_dir = Config.sorted_screenshots_dir.joinpath(new_subdir)

    if is_empty(chosen_filename):
        raise ValueError("Filename can't be blank!")

    if not destination_dir.exists():
        result = psg.popup_yes_no(f"Subdir '{new_subdir}' doesn't exist. Create?",  title="Unknown Subdirectory")

        if result == 'Yes' and not Config.dry_run:
            log.info(f"Creating directory '{new_subdir}'...")
            destination_dir.mkdir()
        else:
            console.print(bullet_text(f"Directory not found. Skipping '{image.file_path}'..."))
            return

    new_filename = destination_dir.joinpath(chosen_filename)
    log.info(f"Chosen Filename: '{chosen_filename}'\nSubdir: '{new_subdir}'\nNew file: '{new_filename}'\nEvent: {event}\n")
    console.print(bullet_text(f"Moving '{image.file_path}' to '{new_filename}'..."))
    image.copy_file_to_sorted_dir(new_filename)
    image.move_to_processed_dir()



"""
Entry point for all of the clown_sort scripts.
"""
import shutil
from argparse import Namespace
from glob import glob
from os import environ, getcwd, path
from pathlib import Path
from typing import List, Union

from dotenv import load_dotenv

# load_dotenv() should be called as soon as possible (before parsing local classes) but not for pytest
if not environ.get('INVOKED_BY_PYTEST', False):
    for dotenv_file in [path.join(dir, '.clown_sort') for dir in [getcwd(), Path.home()]]:
        if path.exists(dotenv_file):
            load_dotenv(dotenv_path=dotenv_file)
            break

from clown_sort.util.argument_parser import (parse_text_extraction_args, parse_pdf_page_extraction_args,
     purge_arg_parser)
from clown_sort.config import Config
from clown_sort.files.image_file import ImageFile
from clown_sort.files.pdf_file import PdfFile
from clown_sort.files.sortable_file import SortableFile
from clown_sort.sort_selector import process_file_with_popup
from clown_sort.util.filesystem_helper import (IMAGE_FILE_EXTENSIONS, files_in_dir, is_image,
      is_pdf, set_timestamp_based_on_screenshot_filename)
from clown_sort.util.logging import log, set_log_level
from clown_sort.util.rich_helper import console


def sort_screenshots():
    """Main entry point for sorting screenshots."""
    Config.configure()

    if Config.rescan_sorted:
        _rescan_sorted_screenshots()
        return

    for file_to_sort in screenshot_paths(Config.screenshots_dir):
        if Config.manual_sort:
            if file_to_sort.can_be_presented_in_popup():
                process_file_with_popup(file_to_sort)
            else:
                print(f"'{file_to_sort.file_path}' is not suitable for manual sort, skipping...")
        else:
            file_to_sort.sort_file()


def extract_text_from_files() -> None:
    """
    Extract text from a single file or from all files in a given directory. Can accept
    multiple paths as arguments on the command line.
    """
    args: Namespace = parse_text_extraction_args()
    console.line()

    if args.debug:
        Config.enable_debug_mode()
    if args.print_as_parsed:
        Config.print_as_parsed = True

    for file_path in args.files_to_process:
        sortable_file = build_sortable_file(file_path)

        if isinstance(sortable_file, PdfFile):
            sortable_file.print_extracted_text(page_range=args.page_range)
        else:
            sortable_file.print_extracted_text()

        console.line(2)


def set_screenshot_timestamps_from_filenames():
    """Parse the filenames to reset the file creation timestamps."""
    Config.configure()

    for image in screenshot_paths(Config.screenshots_dir):
        set_timestamp_based_on_screenshot_filename(image.file_path)


def _rescan_sorted_screenshots():
    """Rescan sorted folders."""
    console.print(f"Rescanning '{Config.sorted_screenshots_dir}'...")
    sortable_files: List[SortableFile] = []
    file_paths: List[str] = []

    for extname in IMAGE_FILE_EXTENSIONS:
        for pattern in ['*', '**/*']:
            glob_pattern = Config.sorted_screenshots_dir.joinpath(f"{pattern}{extname}")
            log.debug(f"Adding '{glob_pattern}' to glob patterns...")
            file_paths.extend(glob(str(glob_pattern)))

    for file_path in file_paths:
        if Config.screenshots_only and not Config.filename_regex.match(path.basename(file_path)):
            log.debug(f"Skipping '{file_path}' because it doesn't match the filename_regex...")
            continue

        sortable_files.append(build_sortable_file(file_path))

    console.print(
        f"Re-processing {len(sortable_files)} files in '{Config.sorted_screenshots_dir}'...",
        style='bright_green'
    )

    for file_path in sortable_files:
        file_path.sort_file()


def screenshot_paths(dir: Path) -> List[SortableFile]:
    """Returns a list of ImageFiles for all the screenshots to be sorted."""
    screenshots = [
        build_sortable_file(f) for f in files_in_dir(Config.screenshots_dir)
        if not Config.screenshots_only or Config.filename_regex.match(path.basename(f))
    ]

    return sorted(screenshots, key=lambda f: f.basename)


def build_sortable_file(file_path: Union[str, Path]) -> SortableFile:
    """Decide if it's a PDF, image, or other type of file."""
    if is_image(file_path):
        return ImageFile(file_path)
    elif is_pdf(file_path):
        return PdfFile(file_path)
    else:
        return SortableFile(file_path)


def purge_non_images_from_dir() -> None:
    """Find all non images in a dir and purge them if they appear elsewhere in the sorted hierarchy."""
    args = Config.configure(purge_arg_parser)
    sorted_files = SortableFile.all_sorted_files()
    set_log_level('INFO')

    for subdir in args.subdirs_to_purge:
        console.print(f"Purging '{subdir}' of non-images...")

        for file_path in files_in_dir(Config.sorted_screenshots_dir.joinpath(subdir)):
            if not is_pdf(file_path):
                log.debug(f"Skipping image '{file_path}'...")
                continue

            basename = path.basename(file_path)
            console.print(f"Checking for '{basename}' in sorted files...")
            matching_files = [f for f in sorted_files if f.name == basename]

            if len(matching_files) <= 1:
                console.print(f" -> Only {len(matching_files)} copies of '{basename}'...", style="dim")
                continue

            processed_file_path = Config.processed_screenshots_dir.joinpath(basename)
            shutil.move(file_path, processed_file_path)
            console.print(f"    Found {len(matching_files)} copies of '{basename}'...")
            console.print(f"    Moved '{file_path}' to '{processed_file_path}'...", style="red")
            console.line()



"""
A range of page numbers.
"""
import re
from argparse import ArgumentTypeError
from dataclasses import dataclass
from typing import Tuple

PAGE_RANGE_REGEX = re.compile('\\d(-\\d)?')


@dataclass
class PageRange:
    page_range: str

    def __post_init__(self):
        if not PAGE_RANGE_REGEX.match(self.page_range):
            raise ValueError(f"Invalid page range '{self.page_range}'")

        if '-' in self.page_range:
            (self.first_page, self.last_page) = (int(p) for p in self.page_range.split('-'))
        else:
            self.first_page = int(self.page_range)
            self.last_page = self.first_page + 1

        if self.last_page <= self.first_page:
            raise ValueError(f"Invalid page range {self.__repr__()}")

    def in_range(self, page_number) -> bool:
        """Returns True if page_number is in this range."""
        return page_number >= self.first_page and page_number < self.last_page

    def file_suffix(self) -> str:
        """String that can be used as file suffix."""
        if self.first_page + 1 == self.last_page:
            return f"page_{self.first_page}"
        else:
            return f"pages_{self.first_page}-{self.last_page}"

    def to_tuple(self) -> Tuple[int, int]:
        return (self.first_page, self.last_page)

    def __repr__(self) -> str:
        return f"PageRange({self.first_page}, {self.last_page})"


class PageRangeArgumentValidator(object):
    HELP_MSG = "a single digit ('11') or a range ('11-15') (WILL NOT extract the last page)"

    def __call__(self, value):
        if not PAGE_RANGE_REGEX.match(value):
            raise ArgumentTypeError("Argument has to match '{}'".format(PAGE_RANGE_REGEX.pattern))

        return PageRange(value)



"""
Logging and printing, for now.
"""
import logging

from rich.logging import RichHandler

from clown_sort.util.constants import PACKAGE_NAME


log = logging.getLogger(PACKAGE_NAME)
log.addHandler(RichHandler(rich_tracebacks=True))


def set_log_level(log_level) -> None:
    log.setLevel(log_level)

    for handler in log.handlers:
        handler.setLevel(log_level)



"""
Base class for sortable files of any type.
"""
import re
import platform
import shutil
from glob import glob
from os import path, remove
from pathlib import Path
from subprocess import run
from typing import List, Optional, Union

from exiftool import ExifToolHelper
from rich.console import Console, ConsoleOptions, RenderResult
from rich.panel import Panel
from rich.prompt import Confirm
from rich.text import Text
from unidecode import unidecode

from clown_sort.config import Config
from clown_sort.filename_extractor import FilenameExtractor
from clown_sort.lib.rule_match import RuleMatch
from clown_sort.sort_selector import process_file_with_popup
from clown_sort.util.filesystem_helper import copy_file_creation_time, loggable_filename
from clown_sort.util.logging import log
from clown_sort.util.rich_helper import (bullet_text, comma_join, console,
     copying_file_log_message, indented_bullet, mild_warning, moving_file_log_message,
     print_dim_bullet, stderr_console)

MAX_EXTRACTION_LENGTH = 4096
NOT_MOVING_FILE = "Not moving file to processed dir because it's"
NO_SORT_FOLDERS_MSG = bullet_text('No sort folders matched so copying to base sorted dir...', style='color(209)')


class SortableFile:
    def __init__(self, file_path: Union[str, Path]) -> None:
        self.file_path: Path = Path(file_path)
        self.basename: str = path.basename(file_path)
        self.basename_without_ext: str = str(Path(self.basename).with_suffix(''))
        self.extname: str = self.file_path.suffix
        self.text_extraction_attempted: bool = False

        self._extracted_text: Optional[str] = None
        self._new_basename: Optional[str] = None
        self._filename_extractor: Optional[FilenameExtractor] = None
        self._paths_of_sorted_copies: List[Path] = []

    def sort_file(self) -> None:
        """Sort the file to destination_dir subdir based on the filename and any extracted text."""
        console.print(self)
        search_text = unidecode(self.basename_without_ext + ' ' + (self.extracted_text() or ''))
        rule_matches = RuleMatch.get_rule_matches(search_text)
        sort_folders = [rm.folder for rm in rule_matches]

        # Handle the case where there are no matches to any configured folders.
        if len(rule_matches) == 0:
            if Config.manual_fallback:
                console.print(Panel('Extracted Text', expand=False))
                console.print(self._extracted_text_panel())

                if self.can_be_presented_in_popup():
                    process_file_with_popup(self)
                    return

                print_dim_bullet(f"'{self.file_path}' cannot be displayed in a popup window yet.")
            elif Config.only_if_match:
                print_dim_bullet('No folder match and --only-if-match option selected. Skipping...')
                return
            elif Config.sorted_screenshots_dir in self.file_path.parents:
                print_dim_bullet("Not moving because no folder match and file already in a sorted folder...")
                return

            console.print(NO_SORT_FOLDERS_MSG)
            sort_folders = [None]
            rule_matches = [None]
        else:
            console.print(bullet_text(Text('Sort folders: ') + comma_join(sort_folders, 'sort_folder')))

        # Copy the renamed file to all the folders whose sorting rules were matched.
        for rule_match in rule_matches:
            # Create the subdir if it doesn't exist.
            folder = None
            match = None

            if rule_match is not None:
                folder = rule_match.folder
                match = rule_match.match
                destination_dir = Config.sorted_screenshots_dir.joinpath(folder)

                if not destination_dir.is_dir() and not Config.dry_run:
                    log.info(f"Creating subdirectory '{destination_dir}'...")
                    destination_dir.mkdir()

            destination_path = self.sort_destination_path(folder)

            if destination_path == self.file_path:
                mild_warning("Source and destination file are the same! Skipping...")
                continue
            elif destination_path.exists():
                if Config.rescan_sorted:
                    mild_warning(f"'{destination_path.name}' already exists in {folder}, skipping...")
                    continue

                if not SortableFile.confirm_file_overwrite(destination_path):
                    continue

            self._paths_of_sorted_copies.append(destination_path)
            self.copy_file_to_sorted_dir(destination_path, match)

        self.move_to_processed_dir()

    def move_to_processed_dir(self) -> None:
        """Finalize the file handling, either leaving, deleting, or moving to processed files dir."""
        if Config.leave_in_place:
            console.print(bullet_text(Text('Leaving in place...', style='dim')))
            return

        # Don't move the file to the processed_dir if it started in a sorted location
        if self.file_path in self._paths_of_sorted_copies:
            console.print(bullet_text(Text('Not moving original file to processed dir...', style='color(127)')))
            return

        if Config.delete_originals:
            self._delete_original()
            return

        self._move_to_processed_dir()

    def new_basename(self) -> str:
        """Basename to rename this file to. Overridden in ImageFile for images with OCR text."""
        return self.basename

    def extracted_text(self) -> Optional[str]:
        """Only PdfFiles and ImageFiles have extracted text; other files are sorted on filename."""
        return self.basename

    def exif_dict(self) -> dict:
        """Return the EXIF data as a dict."""
        try:
            with ExifToolHelper() as exiftool:
                return exiftool.get_metadata(self.file_path)[0]
        except:
            log.warning("ExifTool not found; EXIF data ignored. 'brew install exiftool' may solve this.")
            return {}

    def copy_file_to_sorted_dir(self, destination_path: Path, match: Optional[re.Match] = None):
        """Move or copy the file to destination_subdir."""
        if self.file_path == destination_path:
            console.print(indented_bullet("Source and destination are the same..."))
            return

        self._log_copy_file(destination_path, match)

        if Config.dry_run:
            console.print(indented_bullet("Dry run so not actually copying...", style='dim'))
        else:
            shutil.copy2(self.file_path, destination_path)
            copy_file_creation_time(self.file_path, destination_path)

    def sort_destination_path(self, subdir: Optional[Union[Path, str]] = None) -> Path:
        """Get the destination folder."""
        destination_path = Config.sorted_screenshots_dir

        if subdir is not None:
            destination_path = destination_path.joinpath(subdir)

        return destination_path.joinpath(self.new_basename())

    def preview(self) -> None:
        """Attempt to open a separate application to view the image (Preview on macOS)."""
        log.info(f"Opening '{self.file_path}'")

        if platform.system() == 'Windows':
            log.debug("Windows platform detected; attempting to run the file itself...")
            run([self.file_path])
        else:
            run(['open', self.file_path])

    def print_extracted_text(self) -> None:
        """Pretty print the filename and extracted text."""
        console.print(self._filename_panel())
        console.print(self._extracted_str())

    def can_be_presented_in_popup(self) -> bool:
        """Return True if file can be presented in a popup window for manual sorting (overriden in subclasses)."""
        return False

    def file_size(self) -> int:
        """Returns file size in bytes."""
        return self.file_path.stat().st_size

    def _extracted_str(self, max_chars: Optional[int] = None) -> str:
        """Raw string version of extracted text but truncated to max_chars if provided."""
        txt = self.extracted_text()

        if txt is None:
            txt = "<No extracted text>"
        else:
            if max_chars is not None:
                txt = txt[0:MAX_EXTRACTION_LENGTH]

        return txt

    def _extracted_text_panel(self) -> Panel:
        """Panelized version of the extracted text for display."""
        return Panel(self._extracted_str(MAX_EXTRACTION_LENGTH), expand=True, style='dim')

    def _filename_panel(self) -> Panel:
        """Panelized version of the filename for display."""
        filename = loggable_filename(self.file_path, Config)
        return Panel(filename, expand=False, style='bright_white reverse')

    def _log_copy_file(self, destination_path: Path, match: Optional[re.Match] = None) -> None:
        """Log info about a file copy."""
        if Config.debug:
            console.print(copying_file_log_message(self.basename, destination_path))
            return

        log_msg = Text('').append('Copying to ', style='dim')

        if destination_path.parent == Config.destination_dir:
            console.print(indented_bullet(log_msg.append('root sorted dir...')))
            return

        dirname = loggable_filename(destination_path.parent, Config)
        log_msg.append(str(dirname), style='sort_destination')

        if match is not None:
            log_msg.append(f" (matched '", style='dim')
            log_msg.append(match.group(0).strip(), style='magenta dim')
            log_msg.append("')", style='dim')

        console.print(indented_bullet(log_msg))

    def _move_to_processed_dir(self) -> None:
        """Relocate the original file to the [SCREENSHOTS_DIR]/Processed/ folder."""
        processed_file_path = Config.processed_screenshots_dir.joinpath(self.file_path.name)

        if Config.debug:
            console.print(moving_file_log_message(str(self.file_path), processed_file_path))
        else:
            console.print(bullet_text("Moving to processed dir..."))

        if self.file_path == processed_file_path:
            console.print(indented_bullet(f"{NOT_MOVING_FILE} the same location...", style='dim'))
            return
        elif Config.dry_run:
            msg = f"{NOT_MOVING_FILE} a dry run or --leave-in-place specified..."
            console.print(indented_bullet(msg, style='dim'))
        else:
            shutil.move(self.file_path, processed_file_path)

    def _delete_original(self) -> None:
        """Delete the original file (unless it's a dry run)."""
        console.print(bullet_text(Text(f"Deleting original file...")))

        if Config.dry_run:
            console.print(indented_bullet(Text('Skipping delete because this is a dry run...', style='dim')))
            return

        remove(self.file_path)

    def __str__(self) -> str:
        return str(self.file_path)

    def __repr__(self) -> str:
        return f"SortableFile('{self.file_path}')"

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        """Rich text method."""
        yield Text("\n\n")
        yield self._filename_panel()

        if Config.debug:
            yield self._extracted_text_panel()

        if self._filename_extractor is not None:
            if self._filename_extractor._is_tweet():
                log_txt = bullet_text("It's a tweet by ", style='social_media')
                log_txt.append(self._filename_extractor.author or 'NO_AUTHOR', style='author')

                if self._filename_extractor.reply_to_account is not None:
                    log_txt.append("\n    -> Replying to ", style='color(23)')
                    log_txt.append(self._filename_extractor.reply_to_account, style='author')

                yield log_txt
            elif self._filename_extractor._is_reddit():
                yield Text("It's a reddit post", style='social_media')

        log_msg = Text('Destination filename: ').append(self.new_basename(), style='cyan')
        yield bullet_text(log_msg)

        if Config.debug:
            yield bullet_text('EXIF: ')
            yield f"   {self.exif_dict()}\n\n"

    @classmethod
    def all_sorted_files(cls) -> List[Path]:
        """Return all the files in the sorted directory."""
        file_paths = []

        for pattern in ['*', '**/*']:
            glob_pattern = Config.sorted_screenshots_dir.joinpath(pattern)
            file_paths.extend(glob(str(glob_pattern)))

        return [Path(f) for f in file_paths]

    @staticmethod
    def confirm_file_overwrite(file_path: Path) -> bool:
        """Check if a path exists when about to write to it and ask for confirmation if it does."""
        if not file_path.exists() or Config.yes_overwrite:
            return True

        msg = Text('').append(f"\nWARNING", style='bright_yellow').append(f": File ")
        msg.append(file_path.name, style='cyan').append(" already exists in ")
        msg.append(str(file_path.parent), style='sort_folder')
        stderr_console.print(msg)

        if Config.rescan_sorted:
            stderr_console.print(f"--rescan-sorted flag is on; skipping...", style='dim')
            return False
        elif Confirm.ask(f"Overwrite?", console=stderr_console):
            return True
        else:
            stderr_console.print("Skipping...", style='dim')
            return False



"""
Wrapper for PDF files.
"""
import io
from pathlib import Path
from typing import List, Optional

from pdfalyzer.decorators.pdf_file import PdfFile as PdfalyzerFile

from clown_sort.config import Config, check_for_pymupdf
from clown_sort.files.sortable_file import SortableFile
from clown_sort.lib.page_range import PageRange
from clown_sort.util.constants import PDF_ERRORS
from clown_sort.util.logging import log
from clown_sort.util.rich_helper import WARNING, console, print_error

DEFAULT_PDF_ERRORS_DIR = Path.cwd().joinpath(PDF_ERRORS)
MAX_DISPLAY_HEIGHT = 600
SCALE_FACTOR = 0.4


class PdfFile(SortableFile):
    """
    Wrapper for PDF files.

    Attributes:
        text_extraction_attempted (bool): Whether text extraction has been attempted.
        _extracted_text (Optional[str]): The extracted text from the PDF.
        _page_numbers_of_errors (List[int]): List of page numbers where errors occurred during extraction.
        _is_presentable_in_popup (Optional[bool]): `[class variable]` Cached value indicating if the PDF
            can be presented in a popup.
    """

    _is_presentable_in_popup = None

    def extracted_text(self, page_range: Optional[PageRange] = None) -> Optional[str]:
        """Use PyPDF to extract text page by page and use Tesseract to OCR any embedded images."""
        if self.text_extraction_attempted:
            return self._extracted_text


        pdf_file = PdfalyzerFile(self.file_path)
        self._extracted_text = pdf_file.extract_text(page_range, log, Config.print_as_parsed)
        self.text_extraction_attempted = True
        return self._extracted_text

    def thumbnail_bytes(self) -> Optional[bytes]:
        """Return bytes for a thumbnail image."""
        import fitz  # TODO: Can we do this without PyMuPDF dependency?

        try:
            doc = fitz.open(self.file_path)
        except fitz.EmptyFileError:
            log.warning(f"EmptyFileError: Failed to get bytes for '{self.file_path}'")
            return None

        try:
            # Attempt to resize the thumbnail to fit the screen
            log.debug(f"Getting bytes for '{self.file_path}'...")
            zoom_matrix = fitz.Matrix(fitz.Identity).prescale(SCALE_FACTOR, SCALE_FACTOR)
            page = doc[0]
        except IndexError as e:
            print_error(f"Error getting thumbnail for PDF file '{self.file_path}': {e}")
            return None

        bottom_right = page.rect.br
        page_height = bottom_right[1]
        page_width = bottom_right[0]

        # Check for PDFs with very long pages and crop them
        if SCALE_FACTOR * page_height > MAX_DISPLAY_HEIGHT:
            log.debug(f"PDF page is {page_height} pixels high so cropping...")
            clip = fitz.Rect((0, 0), (MAX_DISPLAY_HEIGHT / SCALE_FACTOR, page_width))
        else:
            clip = fitz.Rect((0, 0), bottom_right)

        return page.get_pixmap(matrix=zoom_matrix, clip= clip, alpha=False).tobytes()

    def print_extracted_text(self, page_range: Optional[PageRange] = None) -> None:
        console.print(self._filename_panel())
        console.print(self.extracted_text(page_range=page_range))

    def can_be_presented_in_popup(self) -> bool:
        """A PDF can be presented in a popup window if PyMuPDF is installed."""
        if type(self)._is_presentable_in_popup is None:
            type(self)._is_presentable_in_popup = check_for_pymupdf()

        if not type(self)._is_presentable_in_popup:
            console.line()
            msg = WARNING.append(f"File '{self.basename}' is not displayable without pymupdf...\n")
            console.print(msg)

        return bool(type(self)._is_presentable_in_popup)

    def __repr__(self) -> str:
        return f"PdfFile('{self.file_path}')"
